import torch

from memoria.engram import Engrams, EngramType
from memoria.memoria import Memoria


def test_add_working_memory():
    memoria = Memoria(
        num_reminded_stm=10,
        ltm_search_depth=3,
        stm_capacity=100,
        initial_lifespan=100,
        num_final_ltms=100,
    )
    memoria.add_working_memory(torch.randn(3, 10, 32))
    assert len(memoria.engrams) == 30


def test_calculate_wm_stm_weight():
    memoria = Memoria(
        num_reminded_stm=10,
        ltm_search_depth=3,
        stm_capacity=100,
        initial_lifespan=100,
        num_final_ltms=100,
    )
    memoria.add_working_memory(torch.randn(3, 10, 32))

    wm = Engrams(torch.randn(3, 10, 32))
    stm = Engrams(torch.randn(3, 20, 32), engrams_types=EngramType.SHORTTERM)
    weight = memoria._calculate_memory_weight(wm, stm)
    assert weight.shape == torch.Size([3, 10, 20])


def test_remind_shortterm_memory():
    memoria = Memoria(
        num_reminded_stm=2,
        ltm_search_depth=3,
        stm_capacity=100,
        initial_lifespan=100,
        num_final_ltms=100,
    )

    weight = torch.tensor([[[0.51, 0.2, 0.2, 0.8]]])
    shortterm_memory_indices = torch.tensor([[1, 2, 3, 4]])
    reminded = memoria._remind_shortterm_memory(weight, shortterm_memory_indices)
    assert (reminded == torch.tensor([[1, -1, -1, 4]])).all()


def test_find_initial_ltm():
    num_stm = 5
    num_ltm = 4
    memoria = Memoria(
        num_reminded_stm=10,
        ltm_search_depth=3,
        stm_capacity=100,
        initial_lifespan=100,
        num_final_ltms=100,
    )

    stm = Engrams(torch.randn(1, num_stm, 32), engrams_types=EngramType.SHORTTERM)
    ltm = Engrams(torch.randn(1, num_ltm, 32), engrams_types=EngramType.LONGTERM)
    engrams = stm + ltm
    memoria.engrams = engrams
    memoria.engrams.induce_counts[:, :num_stm, num_stm:] = torch.tensor(
        [
            [
                [1, 1, 2, 1],
                [1, 1, 1, 2],
                [1, 10, 1, 1],
                [1, 1, 2, 1],
                [1, 5, 1, 1],
            ]
        ]
    )
    memoria.engrams.induce_counts[:, :num_stm, :num_stm] = 999
    nearest_stm_indices = torch.tensor([[0, 2, 3]])

    initial_ltm_indices = memoria._find_initial_longterm_memory(nearest_stm_indices)
    assert (initial_ltm_indices == torch.tensor([[6, 7]])).all()


def test_search_longterm_memories_with_initials():
    num_stm = 5
    num_ltm = 4
    ltm_search_depth = 3
    memoria = Memoria(
        num_reminded_stm=10,
        ltm_search_depth=ltm_search_depth,
        stm_capacity=100,
        initial_lifespan=100,
        num_final_ltms=100,
    )

    stm = Engrams(torch.randn(1, num_stm, 32), engrams_types=EngramType.SHORTTERM)
    ltm = Engrams(torch.randn(1, num_ltm, 32), engrams_types=EngramType.LONGTERM)
    memoria.engrams = stm + ltm

    initial_ltm_indices = torch.tensor([[5, 7]])
    searched_ltm_indices = memoria._search_longterm_memories_with_initials(initial_ltm_indices, ltm)

    assert (searched_ltm_indices == torch.tensor([[1, 3, 2, -1, 0]])).all()


def test_memorize_working_memory_as_shortterm_memory():
    batch_size = 3
    num_wm = 5
    num_stm = 4
    num_ltm = 2
    ltm_search_depth = 3
    memoria = Memoria(
        num_reminded_stm=10,
        ltm_search_depth=ltm_search_depth,
        stm_capacity=100,
        initial_lifespan=100,
        num_final_ltms=100,
    )

    wm = Engrams(torch.randn(batch_size, num_wm, 32), engrams_types=EngramType.WORKING)
    stm = Engrams(torch.randn(batch_size, num_stm, 32), engrams_types=EngramType.SHORTTERM)
    ltm = Engrams(torch.randn(batch_size, num_ltm, 32), engrams_types=EngramType.LONGTERM)
    memoria.engrams = wm + stm + ltm

    memoria._memorize_working_memory_as_shortterm_memory()

    assert memoria.engrams.get_shortterm_memory()[0].data.shape == torch.Size([batch_size, num_wm + num_stm, 32])


def test_memorize_shortterm_memory_as_longterm_memory_or_drop():
    batch_size = 1
    num_stm = 5
    num_ltm = 3
    ltm_search_depth = 3
    memoria = Memoria(
        num_reminded_stm=10,
        ltm_search_depth=ltm_search_depth,
        stm_capacity=2,
        initial_lifespan=100,
        num_final_ltms=100,
    )

    fire_count = torch.tensor([[0, 1, 2, 3, 0]], dtype=torch.int32)
    stm = Engrams(torch.randn(batch_size, num_stm, 32), engrams_types=EngramType.SHORTTERM)
    stm.fire_count = fire_count
    ltm = Engrams(torch.randn(batch_size, num_ltm, 32), engrams_types=EngramType.LONGTERM)
    memoria.engrams = stm + ltm

    memoria._memorize_shortterm_memory_as_longterm_memory()

    assert len(memoria.engrams) == batch_size * (num_stm + num_ltm)


def test_remind():
    num_reminded_stm = 2
    ltm_search_depth = 3
    stm_capacity = 100
    memoria = Memoria(
        num_reminded_stm=num_reminded_stm,
        ltm_search_depth=ltm_search_depth,
        stm_capacity=stm_capacity,
        initial_lifespan=100,
        num_final_ltms=100,
    )

    batch_size = 3
    memory_length = 50
    hidden_dim = 32
    working_memory = torch.randn(batch_size, memory_length, hidden_dim)
    memoria.add_working_memory(working_memory)
    outputs, indices = memoria.remind()
    memoria.adjust_lifespan_and_memories(indices, torch.ones_like(indices, dtype=float))
    assert len(memoria.engrams.get_shortterm_memory()[0]) == batch_size * memory_length
    assert outputs.size(1) == 0

    working_memory = torch.randn(batch_size, memory_length, hidden_dim)
    memoria.add_working_memory(working_memory)
    outputs, indices = memoria.remind()
    memoria.adjust_lifespan_and_memories(indices, torch.ones_like(indices, dtype=float))
    assert len(memoria.engrams.get_shortterm_memory()[0]) == batch_size * memory_length * 2
    assert outputs.size(1) > 0

    working_memory = torch.randn(batch_size, memory_length, hidden_dim)
    memoria.add_working_memory(working_memory)
    outputs, indices = memoria.remind()
    memoria.adjust_lifespan_and_memories(indices, torch.ones_like(indices, dtype=float))
    assert len(memoria.engrams.get_shortterm_memory()[0]) == batch_size * memory_length * 2
    assert outputs.size(1) > 0


def test_reset_memory():
    memoria = Memoria(
        num_reminded_stm=10,
        ltm_search_depth=3,
        stm_capacity=100,
        initial_lifespan=100,
        num_final_ltms=100,
    )
    memoria.add_working_memory(torch.randn(3, 10, 32))
    memoria.reset_memory()
    assert memoria.engrams == Engrams.empty()



from enum import Enum
from typing import Optional, Tuple, Union

import torch

from .types import EngramConnection, EngramInfo, EngramsInfo


class EngramType(Enum):
    NULL = 0
    WORKING = 1
    SHORTTERM = 2
    LONGTERM = 3


class Engrams:
    """Memory Informations

    Attributes:
        batch_size: batch size of data
        memory_length: the number of engrams per each batch element
        data: real memory data shaped [BatchSize, MemoryLength, HiddenDim]
        induce_counts: induce count of each engram about each engram shaped [BatchSize, MemoryLength, MemoryLength]
        engrams_types: engram types of each engram shaped [BatchSize, MemoryLength]
        lifespan: lifespan of each engram shaped [BatchSize, MemoryLength]
        age: age of each engram shaped [BatchSize, MemoryLength]
        engram_ids: id of each engram shaped [BatchSize, MemoryLength]
            if ids is integer, use the valuse as starting id so that id of each engram in a batch will be unique.
        track_age: whether to track age
        track_engram_id: whether to track engram id
    """

    @torch.no_grad()
    def __init__(
        self,
        data: torch.Tensor,
        induce_counts: Optional[torch.Tensor] = None,
        engrams_types: Optional[Union[torch.Tensor, EngramType]] = None,
        lifespan: Union[torch.Tensor, int] = 0,
        age: Optional[Union[torch.Tensor, int]] = None,
        engram_ids: Optional[Union[torch.Tensor, int]] = None,
        track_age: bool = False,
        track_engram_id: bool = False,
    ) -> None:
        self.batch_size, self.memory_length, self.hidden_dim = data.shape
        self.track_age = track_age
        self.track_engram_id = track_engram_id

        self.data: torch.Tensor = data.detach().float()
        self.induce_counts: torch.Tensor = (
            torch.zeros(
                [self.batch_size, self.memory_length, self.memory_length],
                dtype=torch.int32,
                requires_grad=False,
                device=data.device,
            )
            if induce_counts is None
            else induce_counts.detach().type(torch.int32)
        )
        default_engram_type = engrams_types if isinstance(engrams_types, EngramType) else EngramType.WORKING
        self.engrams_types = (
            torch.full(
                [self.batch_size, self.memory_length],
                default_engram_type.value,
                dtype=torch.uint8,
                requires_grad=False,
                device=data.device,
            )
            if not isinstance(engrams_types, torch.Tensor)
            else engrams_types.detach().type(torch.uint8)
        )
        self.lifespan = (
            torch.full(
                [self.batch_size, self.memory_length],
                lifespan,
                dtype=torch.float32,
                requires_grad=False,
                device=data.device,
            )
            if not isinstance(lifespan, torch.Tensor)
            else lifespan.detach().type(torch.float32)
        )

        if self.track_age:
            self.age = (
                torch.zeros(
                    [self.batch_size, self.memory_length],
                    dtype=torch.int32,
                    requires_grad=False,
                    device=data.device,
                )
                if not isinstance(age, torch.Tensor)
                else age.detach().type(torch.int32)
            )
        else:
            self.age = None

        if self.track_engram_id:
            if isinstance(engram_ids, torch.Tensor):
                self.engram_ids = engram_ids.detach().type(torch.int32)
            else:
                engram_ids = engram_ids or 0
                self.engram_ids = (
                    torch.arange(
                        engram_ids,
                        engram_ids + self.memory_length,
                        dtype=torch.int32,
                        requires_grad=False,
                        device=data.device,
                    )
                    .unsqueeze(0)
                    .repeat(self.batch_size, 1)
                )
        else:
            self.engram_ids = None

    @classmethod
    def empty(cls, track_age: bool = False, track_engram_id: bool = False) -> "Engrams":
        return cls(
            torch.zeros([0, 0, 0], dtype=torch.float32, requires_grad=False),
            track_age=track_age,
            track_engram_id=track_engram_id,
        )

    def __len__(self) -> int:
        return self.lifespan.numel()

    @torch.no_grad()
    def __eq__(self, other: "Engrams") -> bool:
        comparison = (
            (self.data == other.data).all()
            and (self.induce_counts == other.induce_counts).all()
            and (self.engrams_types == other.engrams_types).all()
            and (self.lifespan == other.lifespan).all()
            and self.track_age == other.track_age
            and self.track_engram_id == other.track_engram_id
        )

        if self.track_age:
            comparison = comparison and (self.age == other.age).all()
        if self.track_engram_id:
            comparison = comparison and (self.engram_ids == other.engram_ids).all()
        return comparison

    @torch.no_grad()
    def __add__(self, other: "Engrams") -> "Engrams":
        if len(self) == 0:
            return other
        if len(other) == 0:
            return self

        track_age = self.track_age and other.track_age
        use_id = self.track_engram_id and other.track_engram_id
        if self.track_age != other.track_age:
            print("Warning: track_age is disabled due to inconsistency")
        if self.track_engram_id != other.track_engram_id:
            print("Warning: track_engram_id is disabled due to inconsistency")
        concatenated_data = torch.cat([self.data, other.data], dim=1)
        concatenated_engrams_types = torch.cat([self.engrams_types, other.engrams_types], dim=1)
        concatenated_lifespan = torch.cat([self.lifespan, other.lifespan], dim=1)
        concatentaed_age = torch.cat([self.age, other.age], dim=1) if track_age else None
        concatenated_ids = torch.cat([self.engram_ids, other.engram_ids], dim=1) if use_id else None

        new_memory_length = self.memory_length + other.memory_length
        concatenated_induce_counts = torch.zeros(
            [self.batch_size, new_memory_length, new_memory_length],
            dtype=torch.int32,
            requires_grad=False,
            device=self.data.device,
        )
        concatenated_induce_counts[:, : self.memory_length, : self.memory_length] = self.induce_counts
        concatenated_induce_counts[:, self.memory_length :, self.memory_length :] = other.induce_counts

        return Engrams(
            data=concatenated_data,
            induce_counts=concatenated_induce_counts,
            engrams_types=concatenated_engrams_types,
            lifespan=concatenated_lifespan,
            age=concatentaed_age,
            engram_ids=concatenated_ids,
            track_age=track_age,
            track_engram_id=use_id,
        )

    @torch.no_grad()
    def __iter__(self):
        for i in range(self.batch_size):
            yield Engrams(
                data=self.data[i].unsqueeze(0),
                induce_counts=self.induce_counts[i].unsqueeze(0),
                engrams_types=self.engrams_types[i].unsqueeze(0),
                lifespan=self.lifespan[i].unsqueeze(0),
                age=self.age[i].unsqueeze(0) if self.track_age else None,
                engram_ids=self.engram_ids[i].unsqueeze(0) if self.track_engram_id else None,
                track_age=self.track_age,
                track_engram_id=self.track_engram_id,
            )

    @property
    @torch.no_grad()
    def working_memory_mask(self) -> torch.Tensor:
        return self.engrams_types == EngramType.WORKING.value

    @property
    @torch.no_grad()
    def shortterm_memory_mask(self) -> torch.Tensor:
        return self.engrams_types == EngramType.SHORTTERM.value

    @property
    @torch.no_grad()
    def longterm_memory_mask(self) -> torch.Tensor:
        return self.engrams_types == EngramType.LONGTERM.value

    @property
    @torch.no_grad()
    def fire_count(self) -> torch.Tensor:
        return self.induce_counts.diagonal(dim1=1, dim2=2)

    @fire_count.setter
    @torch.no_grad()
    def fire_count(self, value: torch.Tensor) -> None:
        index = torch.arange(value.size(1), device=value.device, dtype=torch.long, requires_grad=False)
        self.induce_counts[:, index, index] = value

    @torch.no_grad()
    def get_indices_with_mask(self, mask: torch.Tensor) -> torch.Tensor:
        """Get global indices with boolean mask

        Args:
            mask: bool type mask tensor shaped [BatchSize, MemoryLength]
        Return:
            global indices shaped [BatchSize, MaxNumSelectedMems]
                the value -1 means null.
        """
        # [BatchSize, MemoryLength]
        num_memories_per_batch = mask.sum(dim=1)
        max_num_memories = num_memories_per_batch.max()
        if (num_memories_per_batch == max_num_memories).all():
            _, indices = mask.nonzero(as_tuple=True)
            indices = indices.view(self.batch_size, max_num_memories)
            return indices

        valued_mask = (
            torch.arange(self.memory_length, device=mask.device, dtype=torch.long, requires_grad=False).unsqueeze(0)
            * mask
        )
        valued_mask.masked_fill_(~mask, -1)
        sorted_values, _ = torch.sort(valued_mask, dim=1)
        indices = sorted_values[:, -max_num_memories:]
        return indices

    @torch.no_grad()
    def get_mask_with_indices(self, indices: torch.Tensor) -> torch.Tensor:
        """Get mask with indices

        Args:
            indices: indices, negative values ignored shaped [BatchSize, NumIndices]
        Return:
            mask: boolean mask shaped [BatchSize, MemoryLength]
        """
        indices[indices < 0] = -1
        mask = torch.zeros(
            [self.batch_size, self.memory_length + 1], dtype=torch.bool, device=indices.device, requires_grad=False
        )
        index_0 = torch.arange(
            self.batch_size,
            device=indices.device,
            dtype=torch.long,
            requires_grad=False,
        ).unsqueeze(1)
        mask[index_0, indices] = True
        mask = mask[:, :-1]
        return mask

    @torch.no_grad()
    def get_working_memory(self) -> Tuple["Engrams", torch.Tensor]:
        """Get working memory engrams and global indices

        Return:
            memory: working memory engrams
            indices: global indices for memory engrams shaped [BatchSize, NumWMems]
                -1 means null index
        """
        mask = self.working_memory_mask
        indices = self.get_indices_with_mask(mask)
        memory = self.mask_select(mask=mask)
        return memory, indices

    @torch.no_grad()
    def get_shortterm_memory(self) -> Tuple["Engrams", torch.Tensor]:
        """Get shortterm memory engrams and global indices

        Return:
            memory: shortterm memory engrams
            indices: global indices for memory engrams shaped [BatchSize, NumSTMems]
                -1 means null index
        """
        mask = self.shortterm_memory_mask
        indices = self.get_indices_with_mask(mask)
        memory = self.mask_select(mask=mask)
        return memory, indices

    @torch.no_grad()
    def get_longterm_memory(self) -> Tuple["Engrams", torch.Tensor]:
        """Get longterm memory engrams and global indices

        Return:
            memory: longterm memory engrams
            indices: global indices for memory engrams shaped [BatchSize, NumLTMems]
                -1 means null index
        """
        mask = self.longterm_memory_mask
        indices = self.get_indices_with_mask(mask)
        memory = self.mask_select(mask=mask)
        return memory, indices

    @torch.no_grad()
    def get_local_indices_from_global_indices(
        self, partial_mask: torch.Tensor, global_indices: torch.Tensor
    ) -> torch.Tensor:
        """Get local ltm indices from global ltm indices
                -1 means null index, ignored

        Args:
            partial_mask: mask to select sub-engrams shaped [BatchSize, MemoryLength]
            global_indices: indices for this engrams shaped [BatchSize, NumIndices]
        """
        partial_indices = self.get_indices_with_mask(partial_mask)
        partial_engrams = self.select(partial_indices)

        selected_partial_mask_shape = list(partial_mask.shape)
        selected_partial_mask_shape[1] += 1
        selected_partial_mask = torch.full(
            selected_partial_mask_shape, False, requires_grad=False, device=partial_mask.device, dtype=torch.bool
        )
        index_0 = torch.arange(
            self.batch_size, requires_grad=False, device=partial_mask.device, dtype=torch.long
        ).unsqueeze(1)
        selected_partial_mask[index_0, global_indices] = True
        selected_partial_mask[:, -1] = False

        # [BatchSize, NumLTMems]
        local_selected_partial_mask = selected_partial_mask[index_0, partial_indices]
        # [BatchSize, NumUniqueInitialLTMs]
        local_selected_ltm_indices = partial_engrams.get_indices_with_mask(local_selected_partial_mask)

        return local_selected_ltm_indices

    @torch.no_grad()
    def fire_together_wire_together(self, indices: torch.Tensor) -> None:
        """Fire & Wire engrams with indices

        Args:
            indices: global indices of firing engrams shaped [BatchSize, NumIndices]
                -1 means ignore, multiple same indices considered once.
        """
        mask = self.get_mask_with_indices(indices)
        self.induce_counts += (mask.float().unsqueeze(2) @ mask.float().unsqueeze(1)).int()

    @torch.no_grad()
    def mask_select(self, mask: torch.Tensor) -> "Engrams":
        """Select values with mask

        Args:
            mask: mask tensor select only true value indices shaped [BatchSize, MemoryLength]
                which is same as self.data
        """
        indices = self.get_indices_with_mask(mask)
        selected_engrams = self.select(indices)
        return selected_engrams

    @torch.no_grad()
    def select(self, indices: torch.Tensor) -> "Engrams":
        """Select indices of memory parts

        Args:
            indices: index tensor for data shaped [NumIndices] or [BatchSize, NumIndices]
                when the indices is 2d-tensor, it can contains -1,
                -1 means null value. it will be null engram types.
        Return:
            Engrams whose data will be shaped [BatchSize, NumIndices, HiddenDim]
        """
        if len(indices.shape) == 1:
            selected_data = self.data[:, indices]
            selected_induce_counts = self.induce_counts[:, indices][:, :, indices]
            selected_engrams_types = self.engrams_types[:, indices]
            selected_lifespan = self.lifespan[:, indices]
            selected_age = self.age[:, indices] if self.track_age else None
            selected_ids = self.engram_ids[:, indices] if self.track_engram_id else None
        elif len(indices.shape) == 2:
            index_0 = torch.arange(self.batch_size, device=self.induce_counts.device, requires_grad=False)
            selected_data = self.data[index_0.unsqueeze(1), indices]
            selected_induce_counts = self.induce_counts[
                index_0.unsqueeze(1).unsqueeze(2), indices.unsqueeze(2), indices.unsqueeze(1)
            ]
            selected_engrams_types = self.engrams_types[index_0.unsqueeze(1), indices]
            selected_lifespan = self.lifespan[index_0.unsqueeze(1), indices]

            null_indices_mask = indices < 0
            reverse_mask = (~null_indices_mask).float()
            selected_data.masked_fill_(null_indices_mask.unsqueeze(2), 0.0)
            selected_induce_counts.masked_fill_(~(reverse_mask.unsqueeze(2) @ reverse_mask.unsqueeze(1)).bool(), -1)
            selected_engrams_types.masked_fill_(null_indices_mask, EngramType.NULL.value)
            selected_lifespan.masked_fill_(null_indices_mask, -1.0)

            if self.track_age:
                selected_age = self.age[index_0.unsqueeze(1), indices]
                selected_age.masked_fill_(null_indices_mask, -1)
            else:
                selected_age = None

            if self.track_engram_id:
                selected_ids = self.engram_ids[index_0.unsqueeze(1), indices]
                selected_ids.masked_fill_(null_indices_mask, -1)
            else:
                selected_ids = None
        else:
            raise ValueError("indices must be 1d or 2d tensor")
        return Engrams(
            data=selected_data,
            induce_counts=selected_induce_counts,
            engrams_types=selected_engrams_types,
            lifespan=selected_lifespan,
            age=selected_age,
            engram_ids=selected_ids,
            track_age=self.track_age,
            track_engram_id=self.track_engram_id,
        )

    @torch.no_grad()
    def delete(self, indices: torch.Tensor) -> None:
        """Delete selected indices engrams

        Args:
            indices: indices of engrams to be deleted shaped [BatchSize, NumIndices]
        """
        mask = self.get_mask_with_indices(indices)
        preserve_indices = self.get_indices_with_mask(~mask)
        selected = self.select(preserve_indices)

        self.batch_size = selected.batch_size
        self.memory_length = selected.memory_length
        self.data = selected.data
        self.induce_counts = selected.induce_counts
        self.engrams_types = selected.engrams_types
        self.lifespan = selected.lifespan
        self.age = selected.age
        self.engram_ids = selected.engram_ids

    @torch.no_grad()
    def extend_lifespan(self, indices: torch.Tensor, lifespan_delta: torch.Tensor):
        """Extend lifespan of selected indices engrams by lifespan_delta

        Args:
            indices: indices of engrams to extend lifespan shaped [BatchSize, NumIndices]
            lifespan_delta: the extended lifespan which will be added to engrams's lifespan
                shaped [BatchSize, NumIndices]
        """
        indices[indices < 0] = -1
        lifespan_delta_mask = torch.zeros(
            [self.batch_size, self.memory_length + 1], dtype=torch.float32, device=indices.device, requires_grad=False
        )
        index_0 = torch.arange(self.batch_size, device=indices.device, requires_grad=False)
        lifespan_delta_mask[index_0.unsqueeze(1), indices] += lifespan_delta
        self.lifespan += lifespan_delta_mask[:, :-1]

    @torch.no_grad()
    def decrease_lifespan(self) -> None:
        self.lifespan -= 1.0

    @torch.no_grad()
    def increase_age(self) -> None:
        if self.track_age:
            self.age += 1

    @torch.no_grad()
    def status_summary(self) -> list[EngramsInfo]:
        """Get status summary of engrams

        Return:
            engrams_info: list of engrams info for each batch element
        """
        assert self.track_engram_id, "Engram id must be tracked to get status summary"

        engrams_info = []
        for e in self:
            engrams = {}
            edges = {}
            engram_ids_by_type = {
                EngramType.WORKING: [],
                EngramType.SHORTTERM: [],
                EngramType.LONGTERM: [],
            }

            engram_ids = e.engram_ids[0].tolist()
            engram_types = e.engrams_types.squeeze(0).tolist()

            # [MemoryLength, MemoryLength]
            # weight_matrix[i, j] means the probability of firing engram i and j together if i fires
            weight_matrix = (e.induce_counts / e.fire_count.unsqueeze(2)).squeeze(0).cpu().numpy()
            induce_counts = e.induce_counts.squeeze(0).tolist()

            for i in range(e.data.size(1)):
                if engram_types[i] == EngramType.NULL.value:
                    continue
                engram_id = engram_ids[i]
                engram_type = EngramType(engram_types[i])
                engram_ids_by_type[engram_type].append(engram_id)
                lifespan = e.lifespan[0, i].item()
                age = e.age[0, i].item() if e.track_age else None

                outgoings = tuple(
                    EngramConnection(source_id=engram_id, target_id=target_id, weight=weight, cofire_count=cofire)
                    for target_id, type, weight, cofire in zip(
                        engram_ids, engram_types, weight_matrix[i].tolist(), induce_counts[i]
                    )
                    if type != EngramType.NULL.value and engram_id != target_id and weight > 0
                )
                incoming = tuple(
                    EngramConnection(source_id=source_id, target_id=engram_id, weight=value, cofire_count=cofire)
                    for source_id, type, value, cofire in zip(
                        engram_ids, engram_types, weight_matrix[:, i].tolist(), induce_counts[i]
                    )
                    if type != EngramType.NULL.value and engram_id != source_id and value > 0
                )
                for outgoing in outgoings:
                    edges[outgoing.source_id, outgoing.target_id] = outgoing

                engrams[engram_id] = EngramInfo(
                    id=engram_id,
                    type=engram_type.name,
                    lifespan=lifespan,
                    age=age,
                    fire_count=e.fire_count[0, i].item(),
                    outgoings=outgoings,
                    incoming=incoming,
                )
            engrams_info.append(
                EngramsInfo(
                    engrams=engrams,
                    edges=edges,
                    working=tuple(engram_ids_by_type[EngramType.WORKING]),
                    shortterm=tuple(engram_ids_by_type[EngramType.SHORTTERM]),
                    longterm=tuple(engram_ids_by_type[EngramType.LONGTERM]),
                )
            )

        return engrams_info



from . import utils
from .abstractor import Abstractor
from .engram import Engrams, EngramType
from .history_manager import HistoryManager
from .memoria import Memoria
from .sparse_tensor import SparseTensor

__all__ = ["utils", "Abstractor", "Engrams", "EngramType", "HistoryManager", "Memoria", "SparseTensor"]
__version__ = "1.0.0"



from itertools import product
from typing import Iterable, List, Optional, Tuple, Union

import torch


class SparseTensor:
    def __init__(
        self,
        indices: torch.Tensor,
        values: torch.Tensor,
        default_value: Union[int, float] = 0,
        shape: Optional[Tuple[int]] = None,
        device: Optional[torch.device] = None,
    ) -> None:
        self.shape = tuple((indices.max(dim=0).values + 1).tolist()) if shape is None else tuple(shape)
        self.device = device if device is not None else values.device

        self.indices = indices.to(self.device)
        self.values = values
        self.default_value = default_value

    def dim(self) -> int:
        return len(self.shape)

    def __getitem__(self, raw_keys: Union[int, slice, torch.Tensor, Tuple, List]) -> "SparseTensor":
        keys = self.__get_keys_with_raw_keys(raw_keys)
        current_data_indices = self.indices.clone()
        current_data_values = self.values.clone()
        current_shape = self.shape

        # Process Scalar Index
        for dim, dim_keys in reversed(list(enumerate(keys))):
            if not isinstance(dim_keys, int):
                continue

            selected_mask = current_data_indices[:, dim] == dim_keys
            selected_values = current_data_values.masked_select(selected_mask)
            selected_indices = torch.cat([current_data_indices[:, :dim], current_data_indices[:, dim + 1 :]], dim=1)
            selected_indices = selected_indices.masked_select(selected_mask.unsqueeze(1)).view(-1, self.dim() - 1)

            current_data_indices = selected_indices
            current_data_values = selected_values
            keys = keys[:dim] + keys[dim + 1 :]
            current_shape = current_shape[:dim] + current_shape[dim + 1 :]

        # Process Slice Index
        for dim, dim_keys in enumerate(keys):
            if not isinstance(dim_keys, slice):
                continue

            dim_key_values = list(range(dim_keys.start, dim_keys.stop, dim_keys.step))
            selected_mask = (current_data_indices[:, dim : dim + 1] == torch.tensor([dim_key_values])).any(dim=1)
            selected_values = current_data_values.masked_select(selected_mask)
            selected_indices = current_data_indices.masked_select(selected_mask.unsqueeze(1)).view(
                -1, current_data_indices.size(1)
            )
            selected_indices[:, dim] -= dim_keys.start
            selected_indices[:, dim] //= dim_keys.step

            current_data_indices = selected_indices
            current_data_values = selected_values
            keys[dim] = None
            current_shape = current_shape[:dim] + (len(dim_key_values),) + current_shape[dim + 1 :]

        # Process Complex Index
        if current_shape == ():
            return (
                current_data_values
                if current_data_values.numel()
                else torch.tensor(self.default_value, device=self.device, dtype=self.values.dtype)
            )

        if all(x is None for x in keys):
            return SparseTensor(current_data_indices, current_data_values, self.default_value, current_shape)

        new_key_rank = max(key.dim() for key in keys if key is not None)
        keys = [v.view([*[1] * (new_key_rank - v.dim()), *v.shape]) if v is not None else None for v in keys]
        new_key_shape = tuple(max(key.shape[i] for key in keys if key is not None) for i in range(new_key_rank))
        new_indices = []
        new_values = []

        dim_keys = [(dim, key) for dim, key in enumerate(keys) if key is not None]
        live_dims = [dim for dim, _ in dim_keys]
        new_indices_dim = live_dims[0]
        for key_idx in product(*[range(s) for s in new_key_shape if s is not None]):
            selected_mask = torch.ones([current_data_indices.size(0)], dtype=torch.bool, device=self.device)
            for dim, key in dim_keys:
                # Boradcast size 1
                key_idx_boardcast = tuple(idx_ if sz != 1 else 0 for idx_, sz in zip(key_idx, key.shape))
                selected_mask &= current_data_indices[:, dim] == key[key_idx_boardcast]

            selected_indices = current_data_indices.masked_select(selected_mask.unsqueeze(1)).view(
                -1, current_data_indices.size(1)
            )
            selected_indices = torch.cat(
                [
                    selected_indices[:, :new_indices_dim],
                    torch.tensor([key_idx]).expand([selected_indices.size(0), len(key_idx)]),
                ]
                + [
                    selected_indices[:, i : i + 1]
                    for i in range(new_indices_dim + 1, selected_indices.size(1))
                    if i not in live_dims
                ],
                dim=1,
            )
            selected_values = current_data_values.masked_select(selected_mask)
            new_indices.append(selected_indices)
            new_values.append(selected_values)

        new_indices = torch.cat(new_indices, dim=0)
        new_values = torch.cat(new_values, dim=0)
        new_shape = []
        for dim, key in enumerate(keys):
            if dim == new_indices_dim:
                new_shape.extend(new_key_shape)
            elif key is None:
                new_shape.append(current_shape[dim])

        return SparseTensor(new_indices, new_values, self.default_value, new_shape)

    def __setitem__(
        self,
        raw_keys: Union[int, slice, torch.Tensor, Tuple, List],
        value: Union[int, float, torch.Tensor, "SparseTensor"],
    ):
        keys = self.__get_keys_with_raw_keys(raw_keys)
        selected_mask = torch.ones([self.indices.size(0)], dtype=torch.bool, device=self.device)
        tensor_key_shape = []
        for dim, key in list(enumerate(keys)):
            if isinstance(key, int):
                keys[dim] = [key]
            elif isinstance(key, slice):
                keys[dim] = range(key.start, key.stop, key.step)
            elif isinstance(key, torch.Tensor):
                if key.dim() < len(tensor_key_shape):
                    key = key.view(*([1] * (len(tensor_key_shape) - key.dim())), *key.shape)
                elif key.dim() > len(tensor_key_shape):
                    tensor_key_shape = [1] * (key.dim() - len(tensor_key_shape)) + tensor_key_shape
                tensor_key_shape = [max(s, ks) for s, ks in zip(key.shape, tensor_key_shape)]
                continue
            elif key is None:
                keys[dim] = range(self.shape[dim])
                continue
            else:
                raise ValueError("key must be int, slice, torch.Tensor, or None")
            key = torch.tensor([keys[dim]], device=self.device, dtype=torch.int32)
            selected_mask &= (self.indices[:, dim : dim + 1] == key).any(dim=1)

        # Process Tensor Key
        for dim, key in list(enumerate(keys)):
            if isinstance(key, torch.Tensor):
                keys[dim] = key.expand(*tensor_key_shape).flatten()

        tensor_keys = [key for key in keys if isinstance(key, torch.Tensor)]
        tensor_key_dims = [dim for dim, key in enumerate(keys) if isinstance(key, torch.Tensor)]
        if tensor_keys:
            tensor_key_indices = self.indices[:, tensor_key_dims]
            key_indices = torch.tensor(list(zip(*tensor_keys)), device=self.device, dtype=torch.int32)
            selected_mask &= (tensor_key_indices.unsqueeze(2) == key_indices.T.unsqueeze(0)).all(dim=1).any(dim=1)

        # Deleted Selected Values
        self.indices = self.indices.masked_select(selected_mask.logical_not().unsqueeze(1)).view(
            -1, self.indices.size(1)
        )
        self.values = self.values.masked_select(selected_mask.logical_not())

        if isinstance(value, int) and self.default_value == value:
            return

        # Add new values
        new_indices = []
        non_tensor_keys = [key for key in keys if not isinstance(key, torch.Tensor)]
        non_and_first_tensor_keys = non_tensor_keys.copy()
        if tensor_keys:
            non_and_first_tensor_keys.insert(tensor_key_dims[0], zip(*tensor_keys) if tensor_keys else (()))
        for new_index in product(*non_and_first_tensor_keys):
            new_index = list(new_index)
            if tensor_keys:
                for dim, key in zip(tensor_key_dims, new_index.pop(tensor_key_dims[0])):
                    new_index.insert(dim, key)
            new_indices.append(new_index)
        new_indices = torch.tensor(new_indices, device=self.device, dtype=torch.int32)

        if isinstance(value, SparseTensor):
            value = value.to_dense()
        if isinstance(value, int) or value.numel() == 1:
            new_values = torch.tensor([value] * new_indices.size(0), device=self.device, dtype=self.values.dtype)
        elif isinstance(value, torch.Tensor):
            tensor_key_start_dim = tensor_key_dims[0]
            value_shape = (
                [len(keys[dim]) for dim in range(tensor_key_start_dim)]
                + [tensor_key_shape]
                + [len(keys[dim]) for dim in range(tensor_key_start_dim + 1, len(keys)) if dim not in tensor_key_dims]
            )
            new_values = value.expand(*value_shape).flatten()

        self.indices = torch.cat([self.indices, new_indices], dim=0)
        self.values = torch.cat([self.values, new_values], dim=0)

    def __add__(self, other: Union[int, "SparseTensor"]) -> "SparseTensor":
        if isinstance(other, int):
            return SparseTensor.from_tensor(self.to_dense() + other)
        elif isinstance(other, SparseTensor):
            if self.default_value != other.default_value:
                raise ValueError("default_value must be the same")
            if self.shape != other.shape:
                raise ValueError("shape must be the same")

            added = self.clone()
            duplicated_mask = (self.indices.unsqueeze(-1) == other.indices.T).all(dim=1)
            added.values[duplicated_mask.any(dim=1).nonzero().squeeze(-1)] += other.values[
                duplicated_mask.any(dim=0).nonzero().squeeze(-1)
            ]
            exclusive_mask = duplicated_mask.any(dim=0).logical_not().nonzero().squeeze(-1)
            added.indices = torch.cat([added.indices, other.indices[exclusive_mask]], dim=0)
            added.values = torch.cat([added.values, other.values[exclusive_mask]], dim=0)
            return added
        else:
            raise ValueError("other must be int or SparseTensor")

    def unsqueeze(self, dim: int = -1) -> "SparseTensor":
        if dim < 0:
            dim += self.dim()

        indices = torch.cat(
            [
                self.indices[:, :dim],
                torch.zeros([self.indices.size(0), 1], dtype=torch.long, device=self.device),
                self.indices[:, dim:],
            ],
            dim=1,
        )
        return SparseTensor(indices, self.values, self.default_value, self.shape[:dim] + (1,) + self.shape[dim:])

    @classmethod
    def empty(
        cls,
        size: Iterable[int],
        default_value: Union[int, float] = 0,
        dtype: Optional[torch.dtype] = torch.int32,
        device: Optional[torch.device] = None,
    ):
        indices = torch.empty((0, len(size)), dtype=torch.long, device=device)
        values = torch.empty((0,), dtype=dtype, device=device)
        return cls(indices, values, default_value, size)

    @classmethod
    def from_tensor(cls, tensor: torch.Tensor, default_value: Union[int, float] = 0):
        indices = torch.nonzero(tensor != default_value, as_tuple=True)
        values = tensor[indices]
        sparse_tensor = cls(torch.stack(indices, dim=1), values, default_value, tensor.shape)
        return sparse_tensor

    def to_dense(self) -> torch.Tensor:
        tensor = torch.full(self.shape, self.default_value, dtype=self.values.dtype, device=self.device)
        tensor[self.indices.unbind(dim=1)] = self.values
        return tensor

    def diagonal(self, dim1: int = 0, dim2: int = 1) -> torch.Tensor:
        if dim1 < 0:
            dim1 += self.dim()
        if dim2 < 0:
            dim2 += self.dim()
        dim1, dim2 = min(dim1, dim2), max(dim1, dim2)

        mask = self.indices[:, dim1] == self.indices[:, dim2]
        indices = self.indices.masked_select(mask.unsqueeze(1)).view(-1, self.indices.size(1))
        values = self.values.masked_select(mask)
        indices = indices[:, [dim for dim in range(indices.size(1)) if dim != dim1]]
        shape = self.shape[:dim1] + self.shape[dim1 + 1 :]
        return SparseTensor(indices, values, self.default_value, shape)

    def to(self, device: torch.device) -> "SparseTensor":
        return SparseTensor(self.indices.to(device), self.values.to(device), self.default_value, self.shape)

    def tolist(self) -> list:
        return self.to_dense().tolist()

    def clone(self) -> "SparseTensor":
        return SparseTensor(self.indices.clone(), self.values.clone(), self.default_value, self.shape)

    def type(self, dtype: torch.dtype) -> "SparseTensor":
        return SparseTensor(self.indices, self.values.type(dtype), self.default_value, self.shape)

    def __eq__(self, other: Union["SparseTensor", torch.Tensor]) -> bool:
        if isinstance(other, torch.Tensor):
            other = SparseTensor.from_tensor(other, self.default_value)
        if isinstance(other, SparseTensor):
            return (
                self.shape == other.shape
                and self.default_value == other.default_value
                and torch.equal(self.indices, other.indices)
                and torch.equal(self.values, other.values)
            )
        return False

    def __get_keys_with_raw_keys(
        self, raw_keys: Union[int, slice, torch.Tensor, Tuple, List]
    ) -> list[Union[int, slice, torch.Tensor]]:
        if not isinstance(raw_keys, tuple):
            raw_keys = (raw_keys,)
        raw_keys = raw_keys + (None,) * (self.dim() - len(raw_keys))

        keys = list(raw_keys)
        for dim, dim_raw_indices in enumerate(raw_keys):
            if isinstance(dim_raw_indices, int) or dim_raw_indices is None:
                pass
            elif isinstance(dim_raw_indices, slice):
                if dim_raw_indices.start is None and dim_raw_indices.stop is None and dim_raw_indices.step is None:
                    keys[dim] = None
                else:
                    start = dim_raw_indices.start or 0
                    stop = min(dim_raw_indices.stop or self.shape[dim], self.shape[dim])
                    step = dim_raw_indices.step or 1
                    keys[dim] = slice(start, stop, step)
            elif isinstance(dim_raw_indices, torch.Tensor):
                keys[dim] = dim_raw_indices
            elif isinstance(dim_raw_indices, list):
                keys[dim] = torch.tensor(dim_raw_indices, dtype=torch.int32, device=self.device)
            else:
                raise ValueError(f"Unsupported index type {type(keys)}")
        return keys

    def __repr__(self) -> str:
        return f"SparseTensor(shape={self.shape}, default_value={self.default_value}, indices={self.indices}, values={self.values})"



import torch


def super_unique(t: torch.Tensor, dim: int) -> torch.Tensor:
    if t.numel() == 0:
        return t

    min_value = t.min()
    t = t - min_value

    max_value = t.max()
    new_shape = list(t.shape)
    new_shape[dim] = max_value + 1
    unique_t_mask = torch.zeros(new_shape, dtype=torch.bool, device=t.device)
    unique_t_mask.scatter_(dim, t.long(), True)

    k = min(t.size(dim), unique_t_mask.sum(dim).max().item())
    validity, unique_t = unique_t_mask.int().topk(k, dim=dim)
    unique_t += min_value
    unique_t.masked_fill_(~validity.bool(), -1)
    return unique_t



import torch
import torch.nn as nn


class Abstractor(nn.Module):
    """Abstract Module to summarize data"""

    def __init__(self, num_memories: int, hidden_dim: int, feedforward_dim: int) -> None:
        """
        Args:
            num_memories (int): Number of memories to be created
            hidden_dim (int): Hidden dimension of the model
            feedforward_dim (int): Feedforward dimension of the model
        """
        super().__init__()

        w = torch.empty(1, num_memories, hidden_dim)
        nn.init.normal_(w, std=0.02)
        self.query_embeddings = nn.Parameter(w)
        self.key_transform = nn.Linear(hidden_dim, hidden_dim)
        self.value_transform = nn.Linear(hidden_dim, hidden_dim)
        self.feedforward = nn.Linear(hidden_dim, feedforward_dim)
        self.output = nn.Linear(feedforward_dim, hidden_dim)

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        """

        Args:
            hidden_states (torch.Tensor): [Batch, N, HiddenDim]
        Returns:
            torch.Tensor: [Batch, NumMemories, HiddenDim]
        """
        query = self.query_embeddings
        key = self.key_transform(hidden_states)
        # [Batch, N, HiddemDim]
        value = self.value_transform(hidden_states)
        # [Batch, NumMemories, HiddenDim] x [Batch, N, HiddenDim] -> [Batch, NumMemories, N]
        attn = query @ key.transpose(-2, -1)
        attn = attn.softmax(dim=-1)
        attn = attn @ value
        attn = self.feedforward(attn)
        attn = self.output(attn)
        return attn



import torch

from memoria.sparse_tensor import SparseTensor


def test_from_tensor():
    tensor = torch.tensor([[1, 0, 0], [0, 2, 0], [0, 0, 3]], dtype=torch.int32)
    sparse_tensor = SparseTensor.from_tensor(tensor)
    assert sparse_tensor.indices.tolist() == [[0, 0], [1, 1], [2, 2]]
    assert sparse_tensor.values.tolist() == [1, 2, 3]


def test_get_item():
    tensor = torch.tensor([[1, 0, 0], [0, 2, 0], [0, 0, 3]], dtype=torch.int32)
    sparse_tensor = SparseTensor.from_tensor(tensor)

    assert sparse_tensor.indices.tolist() == [[0, 0], [1, 1], [2, 2]]
    assert sparse_tensor.values.tolist() == [1, 2, 3]

    selected = sparse_tensor[0, 0]
    assert isinstance(selected, torch.Tensor)
    assert selected.item() == 1

    selected = sparse_tensor[0, 2]
    assert isinstance(selected, torch.Tensor)
    assert selected.item() == 0

    selected = sparse_tensor[1]
    assert selected.shape == (3,)
    assert selected.indices.tolist() == [[1]]
    assert selected.values.tolist() == [2]
    assert selected.tolist() == [0, 2, 0]

    selected = sparse_tensor[:, 2]
    assert selected.shape == (3,)
    assert selected.indices.tolist() == [[2]]
    assert selected.values.tolist() == [3]
    assert selected.tolist() == [0, 0, 3]

    selected = sparse_tensor[torch.tensor([0, 2])]
    assert selected.shape == (2, 3)
    assert selected.indices.tolist() == [[0, 0], [1, 2]]
    assert selected.values.tolist() == [1, 3]
    assert selected.tolist() == [[1, 0, 0], [0, 0, 3]]

    selected = sparse_tensor[torch.tensor([0, 2]), 2]
    assert selected.shape == (2,)
    assert selected.indices.tolist() == [[1]]
    assert selected.values.tolist() == [3]
    assert selected.tolist() == [0, 3]

    selected = sparse_tensor[torch.tensor([[[0, 2]]])]
    assert selected.shape == (1, 1, 2, 3)
    assert selected.indices.tolist() == [[0, 0, 0, 0], [0, 0, 1, 2]]
    assert selected.values.tolist() == [1, 3]
    assert selected.tolist() == [[[[1, 0, 0], [0, 0, 3]]]]

    selected = sparse_tensor[torch.tensor([[0, 1], [2, 0]]), torch.tensor([[1, 2], [2, 0]])]
    assert selected.shape == (2, 2)
    assert selected.indices.tolist() == [[1, 0], [1, 1]]
    assert selected.values.tolist() == [3, 1]
    assert selected.tolist() == [[0, 0], [3, 1]]

    selected = sparse_tensor[0:2]
    assert selected.shape == (2, 3)
    assert selected.indices.tolist() == [[0, 0], [1, 1]]
    assert selected.values.tolist() == [1, 2]
    assert selected.tolist() == [[1, 0, 0], [0, 2, 0]]

    selected = sparse_tensor[0:1, 1:3]
    assert selected.shape == (1, 2)
    assert selected.indices.tolist() == []
    assert selected.values.tolist() == []
    assert selected.tolist() == [[0, 0]]

    selected = sparse_tensor[0:2, 1:3]
    assert selected.shape == (2, 2)
    assert selected.indices.tolist() == [[1, 0]]
    assert selected.values.tolist() == [2]
    assert selected.tolist() == [[0, 0], [2, 0]]

    selected = sparse_tensor[torch.tensor([0, 2]), 1:3]
    assert selected.shape == (2, 2)
    assert selected.indices.tolist() == [[1, 1]]
    assert selected.values.tolist() == [3]
    assert selected.tolist() == [[0, 0], [0, 3]]


def test_set_item():
    tensor = torch.tensor([[1, 0, 0], [0, 2, 0], [0, 0, 3]], dtype=torch.int32)
    sparse_tensor = SparseTensor.from_tensor(tensor)

    sparse_tensor[0, 0] = torch.tensor(90)
    assert sparse_tensor.indices.tolist() == [[1, 1], [2, 2], [0, 0]]
    assert sparse_tensor.values.tolist() == [2, 3, 90]
    assert sparse_tensor.to_dense().tolist() == [[90, 0, 0], [0, 2, 0], [0, 0, 3]]

    sparse_tensor[0, 0] += 10
    assert sparse_tensor.indices.tolist() == [[1, 1], [2, 2], [0, 0]]
    assert sparse_tensor.values.tolist() == [2, 3, 100]
    assert sparse_tensor.to_dense().tolist() == [[100, 0, 0], [0, 2, 0], [0, 0, 3]]

    sparse_tensor = SparseTensor.from_tensor(tensor)
    sparse_tensor[0, 0] = 10
    assert sparse_tensor.indices.tolist() == [[1, 1], [2, 2], [0, 0]]
    assert sparse_tensor.values.tolist() == [2, 3, 10]
    assert sparse_tensor.to_dense().tolist() == [[10, 0, 0], [0, 2, 0], [0, 0, 3]]

    sparse_tensor[1] = 20
    assert sparse_tensor.indices.tolist() == [[2, 2], [0, 0], [1, 0], [1, 1], [1, 2]]
    assert sparse_tensor.values.tolist() == [3, 10, 20, 20, 20]
    assert sparse_tensor.to_dense().tolist() == [[10, 0, 0], [20, 20, 20], [0, 0, 3]]

    sparse_tensor[:, 2] = 30
    assert sparse_tensor.indices.tolist() == [[0, 0], [1, 0], [1, 1], [0, 2], [1, 2], [2, 2]]
    assert sparse_tensor.values.tolist() == [10, 20, 20, 30, 30, 30]
    assert sparse_tensor.to_dense().tolist() == [[10, 0, 30], [20, 20, 30], [0, 0, 30]]

    sparse_tensor = SparseTensor.from_tensor(tensor)
    sparse_tensor[torch.tensor([0, 2])] = 40
    assert sparse_tensor.indices.tolist() == [[1, 1], [0, 0], [0, 1], [0, 2], [2, 0], [2, 1], [2, 2]]
    assert sparse_tensor.values.tolist() == [2, 40, 40, 40, 40, 40, 40]
    assert sparse_tensor.to_dense().tolist() == [[40, 40, 40], [0, 2, 0], [40, 40, 40]]

    sparse_tensor = SparseTensor.from_tensor(tensor)
    sparse_tensor[torch.tensor([0, 2]), 2] = 50
    assert sparse_tensor.indices.tolist() == [[0, 0], [1, 1], [0, 2], [2, 2]]
    assert sparse_tensor.values.tolist() == [1, 2, 50, 50]
    assert sparse_tensor.to_dense().tolist() == [[1, 0, 50], [0, 2, 0], [0, 0, 50]]

    sparse_tensor = SparseTensor.from_tensor(tensor)
    sparse_tensor[torch.tensor([[[0, 2]]])] = 60
    assert sparse_tensor.indices.tolist() == [[1, 1], [0, 0], [0, 1], [0, 2], [2, 0], [2, 1], [2, 2]]
    assert sparse_tensor.values.tolist() == [2, 60, 60, 60, 60, 60, 60]
    assert sparse_tensor.to_dense().tolist() == [[60, 60, 60], [0, 2, 0], [60, 60, 60]]

    sparse_tensor = SparseTensor.from_tensor(tensor)
    sparse_tensor[torch.tensor([[0, 1], [2, 0]]), torch.tensor([[1, 2], [2, 0]])] = 70
    assert sparse_tensor.indices.tolist() == [[1, 1], [0, 1], [1, 2], [2, 2], [0, 0]]
    assert sparse_tensor.values.tolist() == [2, 70, 70, 70, 70]
    assert sparse_tensor.to_dense().tolist() == [[70, 70, 0], [0, 2, 70], [0, 0, 70]]

    sparse_tensor = SparseTensor.from_tensor(tensor)
    sparse_tensor[torch.tensor([[0, 1], [2, 0]]), torch.tensor([[1, 2], [2, 0]])] = torch.tensor([[70, 80], [90, 100]])
    assert sparse_tensor.indices.tolist() == [[1, 1], [0, 1], [1, 2], [2, 2], [0, 0]]
    assert sparse_tensor.values.tolist() == [2, 70, 80, 90, 100]
    assert sparse_tensor.to_dense().tolist() == [[100, 70, 0], [0, 2, 80], [0, 0, 90]]

    sparse_tensor = SparseTensor.from_tensor(tensor)
    sparse_tensor[0:2] = 80
    assert sparse_tensor.indices.tolist() == [[2, 2], [0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2]]
    assert sparse_tensor.values.tolist() == [3, 80, 80, 80, 80, 80, 80]
    assert sparse_tensor.to_dense().tolist() == [[80, 80, 80], [80, 80, 80], [0, 0, 3]]


def test_diagonal():
    tensor = torch.randn(2, 5, 3, 5)
    sparse_tensor = SparseTensor.from_tensor(tensor)

    assert (tensor.diagonal(dim1=1, dim2=3) == sparse_tensor.diagonal(dim1=1, dim2=3).to_dense()).all()


def test_equals():
    tensor = torch.randn(2, 5, 3, 5)
    sparse_tensor = SparseTensor.from_tensor(tensor)

    assert tensor == sparse_tensor
    assert (tensor == sparse_tensor.to_dense()).all()
    assert sparse_tensor != SparseTensor.from_tensor(torch.randn(2, 5, 3, 5))


def test_add():
    tensor = torch.randint(0, 5, [2, 5, 3, 5])
    tensor2 = torch.randint(0, 5, [2, 5, 3, 5])
    sparse_tensor = SparseTensor.from_tensor(tensor)
    sparse_tensor2 = SparseTensor.from_tensor(tensor2)

    assert (tensor + 1 == (sparse_tensor + 1).to_dense()).all()
    assert (tensor + tensor == (sparse_tensor + sparse_tensor).to_dense()).all()
    assert (tensor + tensor2 == (sparse_tensor + sparse_tensor2).to_dense()).all()


def test_unsqueeze():
    tensor = torch.randn(2, 5, 3, 5)
    sparse_tensor = SparseTensor.from_tensor(tensor)

    assert sparse_tensor.unsqueeze(0).shape == (1, 2, 5, 3, 5)
    assert sparse_tensor.unsqueeze(1).shape == (2, 1, 5, 3, 5)
    assert sparse_tensor.unsqueeze(2).shape == (2, 5, 1, 3, 5)
    assert sparse_tensor.unsqueeze(3).shape == (2, 5, 3, 1, 5)
    assert sparse_tensor.unsqueeze(4).shape == (2, 5, 3, 5, 1)


def test_to():
    tensor = torch.randn(2, 5, 3, 5)
    sparse_tensor = SparseTensor.from_tensor(tensor)

    assert sparse_tensor.to(torch.device("cpu")) == sparse_tensor
    assert (sparse_tensor.to(torch.device("cpu")).to_dense() == tensor).all()



import gzip
import pickle
from collections import defaultdict

from .types import EngramHistory, EngramInfo, EngramsInfo, Firing


class HistoryManager:
    """Managing History of engram summaries.

    Attributes:
        timestep: Current timestep.
        summaries: List of engram summaries.
        engram_creation_times: Dictionary of engram creation times.
        engram_deletion_times: Dictionary of engram deletion times.
        engram_durations: Dictionary of engram durations.
        engram_firing_times: Dictionary of engram firing times.
        engram_firings: Dictionary of engram firings.
        firings_per_time: List of firings per timestep.
        engram_fire_counts: Dictionary of engram fire counts.
        engram_ids: List of engram ids.
        alive_engram_ids: List of alive engram ids.
        deleted_engram_ids: List of deleted engram ids.
    """

    def __init__(self):
        self.summaries: list[EngramsInfo] = []
        self.engram_creation_times: dict[int, int] = {}
        self.engram_deletion_times: dict[int, int] = {}
        self.engram_durations: dict[int, int] = {}
        self.engram_firing_times: dict[int, list[int]] = defaultdict(list)
        self.engram_firings: dict[int, list[Firing]] = defaultdict(list)
        self.firings_per_time: list[list[Firing]] = []

        self.alive_engram_ids: list[int] = []
        self.deleted_engram_ids: list[int] = []

    def __len__(self) -> int:
        return len(self.summaries)

    def __getitem__(self, index: int) -> EngramsInfo:
        return self.summaries[index]

    def save(self, path: str) -> None:
        """Save the history manager to a compressed data file.

        Args:
            path: Path to save the history manager.
        """
        with gzip.open(path, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, path: str) -> "HistoryManager":
        """Load the history manager from a compressed data file.

        Args:
            path: Path to load the history manager.
        Returns:
            HistoryManager: Loaded history manager.
        """
        with gzip.open(path, "rb") as f:
            return pickle.load(f)

    @property
    def timestep(self) -> int:
        """Get the current timestep."""
        return len(self)

    @property
    def engram_ids(self) -> list[int]:
        """Get the list of engram IDs."""
        return list(self.engram_creation_times.keys())

    @property
    def engram_fire_counts(self) -> dict[int, int]:
        """Get the fire counts of the engrams."""
        return {engram_id: len(firings) for engram_id, firings in self.engram_firings.items()}

    @property
    def engram_lastest_alive_timestep(self) -> dict[int, int]:
        """Get the latest alive timestep of the engrams."""
        return {
            engram_id: creation_time + self.engram_durations[engram_id] - 1
            for engram_id, creation_time in self.engram_creation_times.items()
        }

    @property
    def latest_engram_infos(self) -> dict[int, EngramInfo]:
        """Get the latest engram information before dying."""
        last_timestep = self.engram_lastest_alive_timestep
        return {engram_id: self.summaries[last_timestep[engram_id]].engrams[engram_id] for engram_id in self.engram_ids}

    def add_summary(self, summary: EngramsInfo) -> None:
        firings = []
        for engram_id, engram in summary.engrams.items():
            if engram_id not in self.alive_engram_ids:
                self.engram_creation_times[engram_id] = self.timestep
                self.alive_engram_ids.append(engram_id)
            elif (
                engram_id in self.summaries[-1].engrams
                and engram.fire_count > self.summaries[-1].engrams[engram_id].fire_count
            ):
                self.engram_firing_times[engram_id].append(self.timestep)
                firing = Firing(
                    timestep=self.timestep,
                    engram_id=engram_id,
                    lifespan_gain=engram.lifespan - self.summaries[-1].engrams[engram_id].lifespan + 1.0,
                )
                self.engram_firings[engram_id].append(firing)
                firings.append(firing)
        self.firings_per_time.append(firings)

        for engram_id in self.alive_engram_ids:
            if engram_id not in summary.engrams:
                self.engram_deletion_times[engram_id] = self.timestep
                self.deleted_engram_ids.append(engram_id)
        self.alive_engram_ids = list(summary.engrams.keys())
        for engram_id in self.alive_engram_ids:
            self.engram_durations[engram_id] = self.timestep - self.engram_creation_times[engram_id] + 1

        self.summaries.append(summary)

    def inspect(self, engram_id: int) -> EngramHistory:
        """Inspect the history of an engram.

        Args:
            engram_id: Engram ID to inspect.
        Returns:
            EngramHistory: Historical information of the engram.
        """
        creation_time = self.engram_creation_times[engram_id]
        deletion_time = self.engram_deletion_times.get(engram_id)
        duration = self.engram_durations.get(engram_id)
        firing_times = self.engram_firing_times.get(engram_id, [])
        firings = self.engram_firings.get(engram_id, [])
        related_summaries = self.summaries[creation_time:deletion_time]

        return EngramHistory(
            id=engram_id,
            creation_timestep=creation_time,
            deletion_timestep=deletion_time,
            duration=duration,
            firing_times=firing_times,
            firings=firings,
            summaries=related_summaries,
        )



import pytest
import torch

from memoria.engram import Engrams, EngramType


def test_empty():
    empty = Engrams.empty()
    empty2 = Engrams.empty()

    assert len(empty) == 0
    assert empty == empty2

    empty_with_options = Engrams.empty(track_age=True, track_engram_id=True)
    assert empty_with_options.age is not None
    assert empty_with_options.engram_ids is not None


def test_init():
    batch_size = 3
    memory_length = 14
    hidden_dim = 32

    data = torch.randn(batch_size, memory_length, hidden_dim)
    engrams = Engrams(data)

    assert (engrams.fire_count == 0).all()
    assert (engrams.induce_counts == 0).all()
    assert (engrams.engrams_types == EngramType.WORKING.value).all()


def test_equals():
    batch_size = 3
    memory_length = 14
    hidden_dim = 32

    data = torch.randn(batch_size, memory_length, hidden_dim)
    engrams = Engrams(data)
    engrams2 = Engrams(data)

    assert engrams == engrams2

    engrams2.fire_count[0, 0] += 1
    assert engrams != engrams2


def test_engram_ids():
    batch_size = 3
    memory_length = 14
    hidden_dim = 32

    data = torch.randn(batch_size, memory_length, hidden_dim)
    engrams = Engrams(data, track_engram_id=True, engram_ids=0)
    assert (engrams.engram_ids == torch.arange(0, memory_length).unsqueeze(0).repeat(batch_size, 1)).all()

    data2 = torch.randn(batch_size, memory_length, hidden_dim)
    engrams2 = Engrams(data2, track_engram_id=True, engram_ids=14)
    assert (engrams2.engram_ids == torch.arange(14, 14 + memory_length).unsqueeze(0).repeat(batch_size, 1)).all()


def test_add():
    data = torch.tensor([[[0.0], [1.0], [2.0], [3.0]]])
    data2 = torch.tensor([[[4.0], [5.0], [6.0], [7.0]]])
    engrams = Engrams(data)
    engrams2 = Engrams(data2, engrams_types=EngramType.SHORTTERM)

    engram_added = engrams + engrams2

    assert len(engrams) + len(engrams2) == len(engram_added)
    assert (engram_added.data == torch.tensor([[[0.0], [1.0], [2.0], [3.0], [4.0], [5.0], [6.0], [7.0]]])).all()
    assert (
        engram_added.engrams_types
        == torch.tensor([[EngramType.WORKING.value] * 4 + [EngramType.SHORTTERM.value] * 4], dtype=torch.int)
    ).all()


def test_iter():
    batch_size = 3
    memory_length = 14
    hidden_dim = 32

    data = torch.randn(batch_size, memory_length, hidden_dim)
    engrams = Engrams(data)

    for i, engram in enumerate(engrams):
        assert engram.data.shape[0] == 1
        assert (engram.data == data[i]).all()


@pytest.mark.parametrize(
    "data,mask,expected",
    [
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]]]),
            torch.tensor([[True, False, True, False], [True, True, False, False]]),
            torch.tensor([[0, 2], [0, 1]]),
        ),
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]]]),
            torch.tensor([[True, False, True, False], [True, True, True, False]]),
            torch.tensor([[-1, 0, 2], [0, 1, 2]]),
        ),
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]]]),
            torch.tensor([[False, False, False, False], [False, False, False, False]]),
            torch.tensor([[], []]),
        ),
    ],
)
def test_get_indices_with_mask(data: torch.Tensor, mask: torch.Tensor, expected: torch.Tensor):
    engrams = Engrams(data)
    indices = engrams.get_indices_with_mask(mask)

    assert (indices == expected).all()


@pytest.mark.parametrize(
    "data,indices,expected",
    [
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]]]),
            torch.tensor([[0, 2], [0, 1]]),
            torch.tensor([[True, False, True, False], [True, True, False, False]]),
        ),
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]]]),
            torch.tensor([[0, 2, -1], [0, 1, 2]]),
            torch.tensor([[True, False, True, False], [True, True, True, False]]),
        ),
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]]]),
            torch.tensor([[], []], dtype=torch.long),
            torch.tensor([[False, False, False, False], [False, False, False, False]]),
        ),
    ],
)
def test_get_mask_with_indices(data: torch.Tensor, indices: torch.Tensor, expected: torch.Tensor):
    engrams = Engrams(data)
    mask = engrams.get_mask_with_indices(indices)

    assert (mask == expected).all()


@pytest.mark.parametrize(
    "data,indices,selected_data,selected_fire_count,selected_induce_counts,selected_engram_types",
    [
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]]]),
            torch.tensor([1, 3]),
            torch.tensor([[[1.0], [3.0]], [[5.0], [7.0]]]),
            torch.zeros([2, 2]),
            torch.zeros([2, 2, 2]),
            torch.tensor([[EngramType.WORKING.value] * 2] * 2),
        ),
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]]]),
            torch.tensor([[1, 3], [0, 2]]),
            torch.tensor([[[1.0], [3.0]], [[4.0], [6.0]]]),
            torch.zeros([2, 2]),
            torch.zeros([2, 2, 2]),
            torch.tensor([[EngramType.WORKING.value] * 2] * 2),
        ),
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]]]),
            torch.tensor([[1, 3], [0, -1]]),
            torch.tensor([[[1.0], [3.0]], [[4.0], [0.0]]]),
            torch.tensor([[0, 0], [0, -1]]),
            torch.tensor([[[0, 0], [0, 0]], [[0, -1], [-1, -1]]]),
            torch.tensor([[EngramType.WORKING.value] * 2, [EngramType.WORKING.value, EngramType.NULL.value]]),
        ),
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]], [[8.0], [9.0], [10.0], [11.0]]]),
            torch.tensor([[1, 3, 2], [0, -1, -1], [2, 3, -1]]),
            torch.tensor([[[1.0], [3.0], [2.0]], [[4.0], [0.0], [0.0]], [[10.0], [11.0], [0.0]]]),
            torch.tensor([[0, 0, 0], [0, -1, -1], [0, 0, -1]]),
            torch.tensor(
                [
                    [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
                    [[0, -1, -1], [-1, -1, -1], [-1, -1, -1]],
                    [[0, 0, -1], [0, 0, -1], [-1, -1, -1]],
                ]
            ),
            torch.tensor(
                [
                    [EngramType.WORKING.value] * 3,
                    [EngramType.WORKING.value, EngramType.NULL.value, EngramType.NULL.value],
                    [EngramType.WORKING.value, EngramType.WORKING.value, EngramType.NULL.value],
                ]
            ),
        ),
    ],
)
def test_select(
    data: torch.Tensor,
    indices: torch.Tensor,
    selected_data: torch.Tensor,
    selected_fire_count: torch.Tensor,
    selected_induce_counts: torch.Tensor,
    selected_engram_types: torch.Tensor,
):
    engrams = Engrams(data)

    selected_engrams = engrams.select(indices)
    assert (selected_engrams.data == selected_data).all()
    assert (selected_engrams.fire_count == selected_fire_count).all()
    assert (selected_engrams.induce_counts == selected_induce_counts).all()
    assert (selected_engrams.engrams_types == selected_engram_types).all()


def test_get_local_indices_from_global_indices():
    data = torch.tensor([[[1], [2], [3], [4], [5], [6]]])
    engrams_types = torch.tensor([[EngramType.SHORTTERM.value] * 3 + [EngramType.LONGTERM.value] * 3])
    engrams = Engrams(data, engrams_types=engrams_types)

    global_indices = [[0, 4, 5, -1]]
    local_indices = engrams.get_local_indices_from_global_indices(engrams.longterm_memory_mask, global_indices)

    assert (local_indices == torch.tensor([[1, 2]])).all()


def test_fire_together_wire_together():
    data = torch.tensor([[[1], [2], [3], [4], [5], [6]]])
    engrams = Engrams(data)

    engrams.fire_together_wire_together(torch.tensor([[0, 2, 4, -1]]))
    assert (engrams.fire_count == torch.tensor([[1, 0, 1, 0, 1, 0]])).all()
    assert (
        engrams.induce_counts
        == torch.tensor(
            [
                [1, 0, 1, 0, 1, 0],
                [0, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 1, 0],
                [0, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 1, 0],
                [0, 0, 0, 0, 0, 0],
            ]
        )
    ).all()

    engrams.fire_together_wire_together(torch.tensor([[0, 1, 2, 2, 3, -1, -10]]))
    assert (engrams.fire_count == torch.tensor([[2, 1, 2, 1, 1, 0]])).all()
    assert (
        engrams.induce_counts
        == torch.tensor(
            [
                [2, 1, 2, 1, 1, 0],
                [1, 1, 1, 1, 0, 0],
                [2, 1, 2, 1, 1, 0],
                [1, 1, 1, 1, 0, 0],
                [1, 0, 1, 0, 1, 0],
                [0, 0, 0, 0, 0, 0],
            ]
        )
    ).all()


@pytest.mark.parametrize(
    "data,mask,selected_data,selected_fire_count,selected_induce_counts,selected_engram_types",
    [
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]]]),
            torch.tensor([[False, True] * 2] * 2),
            torch.tensor([[[1.0], [3.0]], [[5.0], [7.0]]]),
            torch.zeros([2, 2]),
            torch.zeros([2, 2, 2]),
            torch.tensor([[EngramType.WORKING.value] * 2] * 2),
        ),
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]]]),
            torch.tensor([[False, True, False, True], [True, False, True, False]]),
            torch.tensor([[[1.0], [3.0]], [[4.0], [6.0]]]),
            torch.zeros([2, 2]),
            torch.zeros([2, 2, 2]),
            torch.tensor([[EngramType.WORKING.value] * 2] * 2),
        ),
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]]]),
            torch.tensor([[False, True, False, True], [True, False, False, False]]),
            torch.tensor([[[1.0], [3.0]], [[0.0], [4.0]]]),
            torch.tensor([[0, 0], [-1, 0]]),
            torch.tensor([[[0, 0], [0, 0]], [[-1, -1], [-1, 0]]]),
            torch.tensor([[EngramType.WORKING.value] * 2, [EngramType.NULL.value, EngramType.WORKING.value]]),
        ),
        (
            torch.tensor([[[0.0], [1.0], [2.0], [3.0]], [[4.0], [5.0], [6.0], [7.0]], [[8.0], [9.0], [10.0], [11.0]]]),
            torch.tensor([[False, True, True, True], [True, False, False, False], [False, False, True, True]]),
            torch.tensor([[[1.0], [2.0], [3.0]], [[0.0], [0.0], [4.0]], [[0.0], [10.0], [11.0]]]),
            torch.tensor([[0, 0, 0], [-1, -1, 0], [-1, 0, 0]]),
            torch.tensor(
                [
                    [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
                    [[-1, -1, -1], [-1, -1, -1], [-1, -1, 0]],
                    [[-1, -1, -1], [-1, 0, 0], [-1, 0, 0]],
                ]
            ),
            torch.tensor(
                [
                    [EngramType.WORKING.value] * 3,
                    [EngramType.NULL.value, EngramType.NULL.value, EngramType.WORKING.value],
                    [EngramType.NULL.value, EngramType.WORKING.value, EngramType.WORKING.value],
                ]
            ),
        ),
    ],
)
def test_mask_select(
    data: torch.Tensor,
    mask: torch.Tensor,
    selected_data: torch.Tensor,
    selected_fire_count: torch.Tensor,
    selected_induce_counts: torch.Tensor,
    selected_engram_types: torch.Tensor,
):
    engrams = Engrams(data)

    selected_engrams = engrams.mask_select(mask)
    assert (selected_engrams.data == selected_data).all()
    assert (selected_engrams.fire_count == selected_fire_count).all()
    assert (selected_engrams.induce_counts == selected_induce_counts).all()
    assert (selected_engrams.engrams_types == selected_engram_types).all()


def test_get_memories():
    data = torch.tensor(
        [
            [[1.0], [2.0], [3.0], [4.0], [5.0]],
            [[6.0], [7.0], [8.0], [9.0], [10.0]],
        ]
    )
    NULL = EngramType.NULL.value
    WORKING = EngramType.WORKING.value
    SHORTTERM = EngramType.SHORTTERM.value
    LONGTERM = EngramType.LONGTERM.value
    engrams_types = torch.tensor(
        [
            [WORKING, WORKING, LONGTERM, NULL, SHORTTERM],
            [SHORTTERM, NULL, LONGTERM, LONGTERM, WORKING],
        ]
    )
    engrams = Engrams(data=data, engrams_types=engrams_types)

    wm_engrams, wm_indices = engrams.get_working_memory()
    stm_engrams, stm_indices = engrams.get_shortterm_memory()
    ltm_engrams, ltm_indices = engrams.get_longterm_memory()

    assert wm_engrams == engrams.select(wm_indices)
    assert stm_engrams == engrams.select(stm_indices)
    assert ltm_engrams == engrams.select(ltm_indices)

    assert (engrams.select(wm_indices).data == torch.tensor([[[1.0], [2.0]], [[0.0], [10.0]]])).all()
    assert (engrams.select(stm_indices).data == torch.tensor([[[5.0]], [[6.0]]])).all()
    assert (engrams.select(ltm_indices).data == torch.tensor([[[0.0], [3.0]], [[8.0], [9.0]]])).all()

    assert ((wm_engrams.engrams_types == WORKING) | (wm_engrams.engrams_types == NULL)).all()
    assert ((stm_engrams.engrams_types == SHORTTERM) | (stm_engrams.engrams_types == NULL)).all()
    assert ((ltm_engrams.engrams_types == LONGTERM) | (ltm_engrams.engrams_types == NULL)).all()


def test_delete():
    data = torch.tensor(
        [
            [[1.0], [2.0], [3.0], [4.0], [5.0]],
            [[6.0], [7.0], [8.0], [9.0], [10.0]],
        ]
    )
    engrams = Engrams(data=data)
    engrams.delete(torch.tensor([[1, 3], [4, -1]]))

    assert (
        engrams.data
        == torch.tensor(
            [
                [[0.0], [1.0], [3.0], [5.0]],
                [[6.0], [7.0], [8.0], [9.0]],
            ]
        )
    ).all()
    assert (engrams.fire_count == torch.tensor([[-1, 0, 0, 0], [0, 0, 0, 0]])).all()


def test_lifespan():
    data = torch.tensor(
        [
            [[1.0], [2.0], [3.0], [4.0], [5.0]],
            [[6.0], [7.0], [8.0], [9.0], [10.0]],
        ]
    )
    lifespan = torch.tensor(
        [
            [1.0, 2.0, 3.0, 4.0, 5.0],
            [2.0, 3.0, 4.0, 5.0, 6.0],
        ]
    )
    engrams = Engrams(data=data, lifespan=lifespan)

    indices = torch.tensor(
        [
            [4, -1, 0],
            [-1, 2, -1],
        ]
    )
    lifespan_delta = torch.tensor(
        [
            [1.0, 2.0, 3.0],
            [2.0, 3.0, 4.0],
        ]
    )
    engrams.extend_lifespan(indices, lifespan_delta)

    assert (
        engrams.lifespan
        == torch.tensor(
            [
                [4.0, 2.0, 3.0, 4.0, 6.0],
                [2.0, 3.0, 7.0, 5.0, 6.0],
            ]
        )
    ).all()

    engrams.decrease_lifespan()
    assert (
        engrams.lifespan
        == torch.tensor(
            [
                [3.0, 1.0, 2.0, 3.0, 5.0],
                [1.0, 2.0, 6.0, 4.0, 5.0],
            ]
        )
    ).all()


def test_status_summary():
    data = torch.tensor(
        [
            [[1.0], [2.0], [3.0], [4.0], [5.0]],
        ]
    )
    NULL = EngramType.NULL.value
    WORKING = EngramType.WORKING.value
    SHORTTERM = EngramType.SHORTTERM.value
    LONGTERM = EngramType.LONGTERM.value
    engrams_types = torch.tensor(
        [
            [WORKING, WORKING, LONGTERM, NULL, SHORTTERM],
        ]
    )
    lifespan = torch.tensor([[1.0, 2.0, 3.0, 4.0, 5.0]])
    age = torch.tensor([[0, 1, 2, 3, 4]])
    ids = torch.tensor([[0, 1, 2, 3, 4]])
    induce_counts = torch.tensor(
        [
            [5, 1, 1, 3, 1],
            [1, 4, 1, 0, 4],
            [1, 0, 3, 0, 3],
            [2, 2, 1, 2, 1],
            [1, 2, 3, 0, 7],
        ],
    ).unsqueeze(0)

    engrams = Engrams(
        data=data,
        engrams_types=engrams_types,
        induce_counts=induce_counts,
        lifespan=lifespan,
        age=age,
        engram_ids=ids,
        track_engram_id=True,
        track_age=True,
    )
    info = engrams.status_summary()[0]

    assert info.working == (0, 1)
    assert info.shortterm == (4,)
    assert info.longterm == (2,)
    assert len(info.edges) == 4 * 4 - 4 - 1
    assert len(info.engrams) == 4
    assert info.engrams[2].lifespan == 3.0
    assert info.engrams[0].type == "WORKING"
    assert info.engrams[1].age == 1
    assert [edge.weight for edge in info.engrams[4].outgoings] == (torch.tensor([1, 2, 3]) / 7).tolist()
    assert [edge.weight for edge in info.engrams[1].incoming] == (torch.tensor([1, 2]) / torch.tensor([5, 7])).tolist()
    assert info.engrams[0].outgoings[2].cofire_count == 1
    assert info.engrams[1].outgoings[2].cofire_count == 4
    assert info.engrams[4].fire_count == 7



import sys
from typing import Optional, Tuple, Union

import torch

from .engram import Engrams, EngramType
from .history_manager import HistoryManager
from .utils import super_unique


class Memoria:
    """Memoria, general memory network for sequential modeling or processing

    Attributes:
        engrams: engrams to store memory data and the relations between them
        next_engram_id: next engram id to be assigned
        timestep: current timestep
        history: history of engrams summaries
        num_reminded_stm: the number of stm reminded
        stm_capacity: the maximum memory length per batch for shortterm memory
        ltm_search_depth: the maximum number of depth for dfs memory search
        initial_lifespan: initial lifespan for each engrams
        num_final_ltms: the number of longterm memories to return
        track_age: whether to track age of each engram
        track_id: whether to track id of each engram
        keep_history: whether to keep history of engrams
        device: memoria device to save engrams
        ext_device: original device of the data
    """

    def __init__(
        self,
        num_reminded_stm: float,
        stm_capacity: int,
        ltm_search_depth: int,
        initial_lifespan: int,
        num_final_ltms: int,
        track_age: bool = False,
        track_id: bool = False,
        keep_history: bool = False,
        device: Optional[Union[str, torch.device]] = None,
    ) -> None:
        """Initialize Memoria

        Args:
            num_reminded_stm: the number of stm reminded
            stm_capacity: the maximum memory length per batch for shortterm memory
            ltm_search_depth: the maximum number of depth for dfs memory search
            initial_lifespan: initial lifespan for each engrams
            num_final_ltms: the number of longterm memories to return
                this module will not return shortterm memory indices, but keep shortterm memory
                to remind longterm memory. so when `enable ltm` is True.
            track_age: whether to track age of each engram
            track_id: whether to track id of each engram
            keep_history: whether to keep history of engrams
            device: memoria device to save engrams
        """
        self.engrams = None
        self.next_engram_id = None
        self.timestep = None
        self.history = None

        if keep_history:
            if not track_id:
                track_id = True
                print("Warning: `keep_history` is True, so `track_id` is automatically set to True", file=sys.stderr)
            if not track_age:
                track_age = True
                print("Warning: `keep_history` is True, so `track_age` is automatically set to True", file=sys.stderr)

        self.num_reminded_stm: float = num_reminded_stm
        self.stm_capacity: int = stm_capacity
        self.ltm_search_depth: int = ltm_search_depth
        self.initial_lifespan: int = initial_lifespan
        self.num_final_ltms: int = num_final_ltms
        self.track_age: bool = track_age
        self.track_id: bool = track_id
        self.keep_history: bool = keep_history
        self.device = torch.device(device) if device else None
        self.ext_device = None

        self.reset_memory()

    @torch.no_grad()
    def add_working_memory(self, data: torch.Tensor) -> None:
        """Add new working memories to engrams

        Args:
            data: input data which is working memory shaped [BatchSize, WorkingMemoryLength, HiddenDim]
        """
        self.ext_device = data.device
        self.device = self.device or data.device
        new_engrams = Engrams(
            data.to(self.device),
            engrams_types=EngramType.WORKING,
            lifespan=self.initial_lifespan,
            engram_ids=self.next_engram_id,
            track_age=self.track_age,
            track_engram_id=self.track_id,
        )
        self.engrams += new_engrams

        if self.track_id:
            self.next_engram_id = new_engrams.engram_ids.max().cpu().item() + 1
        self.timestep += 1

    @torch.no_grad()
    def remind(self) -> Tuple[torch.Tensor, torch.Tensor]:
        """Remind with current working memory and memorize, drop memories

        Return:
            reminded indices, data not containing working memory
            data shaped [BatchSize, RemindedMemoryLength, HiddenDim]
            indices shaped [BatchSize, RemindedMemoryLength]
        """
        wm_engrams, wm_indices = self.engrams.get_working_memory()
        stm_engrams, stm_indices = self.engrams.get_shortterm_memory()
        ltm_engrams, _ = self.engrams.get_longterm_memory()

        weight = self._calculate_memory_weight(wm_engrams, stm_engrams)
        reminded_stm_indices = self._remind_shortterm_memory(weight, stm_indices)
        initial_ltm_indices = self._find_initial_longterm_memory(reminded_stm_indices)
        searched_ltm_indices = self._search_longterm_memories_with_initials(initial_ltm_indices, ltm_engrams)
        reminded_ltm_indices = self._select_final_ltms(wm_engrams, searched_ltm_indices)

        reminded_indices = torch.cat([reminded_stm_indices, reminded_ltm_indices], dim=1)

        fire_indices = torch.cat([wm_indices, reminded_indices], dim=1)
        self.engrams.fire_together_wire_together(fire_indices)

        reminded_indices = super_unique(reminded_indices, dim=1)

        # [BatchSize, RemindedMemoryLength, HiddenDim]
        reminded_memories = self.engrams.select(reminded_indices).data.to(self.ext_device).contiguous()

        return reminded_memories, reminded_indices

    @torch.no_grad()
    def adjust_lifespan_and_memories(self, indices: torch.Tensor, lifespan_delta: torch.Tensor):
        """Adjust lifespan and convert working memories into shortterm memories,
            shortterm memories into longterm memories, and drop memories whose lifespan is zero

        Args:
            indices: indices to adjust lifespan shaped [BatchSize, RemindedMemoryLength]
                normally, this is the indices returned from `remind` method
            lifespan_delta: float tensor shaped [BatchSize, RemindedMemoryLength]
                the value of lifespan_delta[i][j] is increment of the engram whose index is indices[i][j]
        """
        lifespan_delta = lifespan_delta.to(self.device)
        self.engrams.extend_lifespan(indices, lifespan_delta)
        self.engrams.decrease_lifespan()
        self.engrams.increase_age()

        self._memorize_working_memory_as_shortterm_memory()
        self._memorize_shortterm_memory_as_longterm_memory()

        self.engrams = self.engrams.mask_select(self.engrams.lifespan > 0)

        if self.keep_history:
            if self.history is None:
                self.history = [HistoryManager() for _ in range(indices.size(0))]
            for history, summary in zip(self.history, self.engrams.status_summary()):
                history.add_summary(summary)

    @torch.no_grad()
    def _calculate_memory_weight(self, working_memory: Engrams, key_memory: Engrams) -> torch.Tensor:
        """Calculate attention weight from working memories over shotterm/longterm memories

        Returns:
            attention weights shaped [BatchSize, WorkingMemoryLength, ShorttermMemoryLength]
        """
        l2_square = (working_memory.data.unsqueeze(2) - key_memory.data.unsqueeze(1)).pow(2).sum(dim=3)
        return l2_square.neg()

    @torch.no_grad()
    def _remind_shortterm_memory(self, weight: torch.Tensor, shortterm_memory_indices: torch.Tensor) -> torch.Tensor:
        """Remind shortterm memories by its working memories

        Args:
            weight: attention weights got from `calculate_wm_stm_weight` method
                shaped [BatchSize, WorkingMemoryLength, ShorttermMemoryLength]
            shortterm_memory_indices: global indices of shortterm memories shaped [BatchSize, ShorttermMemoryLength]
        Returns:
            indices selected shortterm memory indices shaped [BatchSize, SelectedMemoryLength]
            -1 means not reminded, non-negative intergers is reminded
        """
        # [BatchSize, ShorttermMemoryLength]
        stm_weight = weight.mean(dim=1)

        _, reminded_indices = stm_weight.topk(min(self.num_reminded_stm, stm_weight.size(1)), dim=1)
        reminded_mask = torch.zeros_like(
            shortterm_memory_indices, dtype=bool, device=weight.device, requires_grad=False
        )
        index_0 = torch.arange(weight.size(0), device=weight.device, requires_grad=False, dtype=torch.long)
        reminded_mask[index_0.unsqueeze(1), reminded_indices] = True
        return shortterm_memory_indices.masked_fill(~reminded_mask, -1)

    @torch.no_grad()
    def _find_initial_longterm_memory(self, nearest_shortterm_memory_indices: torch.Tensor) -> torch.Tensor:
        """Search nearest longterm memory indices

        Args:
            nearest_shortterm_memory_indices: shortterm memory indices shaped [BatchSize, NumInitialLTMs]
                got from `find_stm_nearest_to_ltm` method
        Return:
            initial longterm memory indices to be reminded shaped [BatchSize, NumInitialLTMs]
        """
        index_0 = torch.arange(
            nearest_shortterm_memory_indices.size(0),
            requires_grad=False,
            device=nearest_shortterm_memory_indices.device,
        ).unsqueeze(1)
        # [BatchSize, NumInitialLTMs, MemoryLength]
        induce_counts = self.engrams.induce_counts[index_0, nearest_shortterm_memory_indices]
        # [BatchSize, MemoryLength]
        ltm_mask = self.engrams.longterm_memory_mask

        induce_counts.masked_fill_(~ltm_mask.unsqueeze(1), -1)

        # [BatchSize, NumInitialLTMs]
        # Find initial ltm indices with random tie-breaking
        max_mask = induce_counts == induce_counts.max(dim=2).values.unsqueeze(dim=2)
        initial_ltm_indices = (torch.rand_like(max_mask.type(torch.float16)) * max_mask).argmax(dim=2)
        initial_ltm = self.engrams.select(initial_ltm_indices)
        initial_ltm_indices.masked_fill_(initial_ltm.engrams_types != EngramType.LONGTERM.value, -1)

        # [BatchSize, NumUniqueInitialLTMs]
        initial_ltm_indices = super_unique(initial_ltm_indices, dim=1)
        return initial_ltm_indices

    @torch.no_grad()
    def _search_longterm_memories_with_initials(
        self, initial_longterm_memory_indices: torch.Tensor, longterm_memory: Engrams
    ) -> torch.Tensor:
        """Find ltm engrams with initila ltm indices by dfs method

        Args:
            initial_longterm_memory_indices: initial ltm indices shaped [BatchSize, NumUniqueInitialLTMs]
            longterm_memory: longterm memory engrams
        Return:
            searched ltm indices shaped [BatchSize, (SearchDepth + 1) * NumUniqueInitialLTMs]
        """
        batch_size, num_init_ltms = initial_longterm_memory_indices.shape
        local_initial_ltm_indices = self.engrams.get_local_indices_from_global_indices(
            self.engrams.longterm_memory_mask, initial_longterm_memory_indices
        )
        if initial_longterm_memory_indices.numel() == 0 or len(longterm_memory) == 0:
            return torch.zeros(
                [batch_size, (self.ltm_search_depth + 1) * num_init_ltms],
                dtype=torch.long,
                device=initial_longterm_memory_indices.device,
                requires_grad=False,
            )

        # [BatchSize, NumLTMems]
        unreachable = ~longterm_memory.longterm_memory_mask
        found_ltm_indices = torch.full(
            [batch_size, self.ltm_search_depth + 1, min(num_init_ltms, local_initial_ltm_indices.size(1))],
            -1,
            requires_grad=False,
            device=local_initial_ltm_indices.device,
            dtype=local_initial_ltm_indices.dtype,
        )

        index_0 = torch.arange(
            batch_size, device=local_initial_ltm_indices.device, requires_grad=False, dtype=torch.long
        ).unsqueeze(1)
        found_ltm_indices[:, 0] = local_initial_ltm_indices
        unreachable[index_0, local_initial_ltm_indices] = True

        for depth in range(self.ltm_search_depth):
            last_ltm_indices = found_ltm_indices[:, depth]
            reachable_induce_counts = longterm_memory.induce_counts.masked_fill(unreachable.unsqueeze(1), -1)
            # [BatchSize, NumUniqueInitialLTMs, NumLTMems]
            last_ltm_reachable_induce_counts = reachable_induce_counts[index_0, last_ltm_indices]

            current_ltm_indices = last_ltm_reachable_induce_counts.argmax(dim=2)
            current_ltm_indices.masked_fill_(unreachable[index_0, current_ltm_indices], -1)
            found_ltm_indices[:, depth + 1] = current_ltm_indices
            unreachable[index_0, current_ltm_indices] = True

        return super_unique(found_ltm_indices.view(batch_size, -1), dim=1)

    @torch.no_grad()
    def _select_final_ltms(self, working_memory: Engrams, found_longterm_memory_indices: torch.Tensor) -> torch.Tensor:
        """Select top k LTM indices by working memory weights

        Args:
            working_memory: working memory engrams
            found_longterm_memory_indices: found ltm indices shaped [BatchSize, NumFoundLTMs]
        Return:
            final ltm indices shaped [BatchSize, NumFinalLTMs]
        """
        if found_longterm_memory_indices.size(1) == 0:
            return found_longterm_memory_indices

        # [BatchSize, NumFoundLTMs]
        found_ltms = self.engrams.select(found_longterm_memory_indices)
        # [BatchSize, NumFoundLTMs]
        weight = self._calculate_memory_weight(working_memory, found_ltms).sum(dim=1)

        topk = min(self.num_final_ltms, weight.size(1))
        # [BatchSize, NumFinalLTMs]
        final_ltm_indices = weight.topk(topk, dim=1).indices
        final_ltm_indices = found_longterm_memory_indices.gather(dim=1, index=final_ltm_indices)
        return final_ltm_indices

    @torch.no_grad()
    def _memorize_working_memory_as_shortterm_memory(self):
        """Move working memory to shortterm memory"""
        self.engrams.engrams_types[self.engrams.working_memory_mask] = EngramType.SHORTTERM.value

    @torch.no_grad()
    def _memorize_shortterm_memory_as_longterm_memory(self):
        """Move exceeded shortterm memory to longterm memory or drop"""
        stm_engrams, stm_indices = self.engrams.get_shortterm_memory()
        num_exceeded_stm = stm_engrams.memory_length - self.stm_capacity

        if num_exceeded_stm <= 0:
            return

        # [BatchSize, NumExceededSTMems]
        exceeded_stm_indices = stm_indices[:, :num_exceeded_stm]
        memorize_stm_mask = self.engrams.get_mask_with_indices(exceeded_stm_indices)
        self.engrams.engrams_types[memorize_stm_mask] = EngramType.LONGTERM.value

    def reset_memory(self):
        """Reset memory"""
        self.engrams = Engrams.empty(track_age=self.track_age, track_engram_id=self.track_id)
        self.next_engram_id = 0
        self.timestep = 0
        self.history = None



import torch

from memoria.abstractor import Abstractor


def test_abstractor():
    abstractor = Abstractor(num_memories=3, hidden_dim=4, feedforward_dim=5)
    hidden_states = torch.randn(2, 3, 4)
    output = abstractor(hidden_states)
    assert output.shape == (2, 3, 4)



from memoria.history_manager import HistoryManager
from memoria.types import EngramInfo, EngramsInfo, Firing


def test_history_manager():
    history_manager = HistoryManager()
    assert len(history_manager) == 0
    assert history_manager.timestep == 0
    assert history_manager.engram_ids == []
    assert history_manager.alive_engram_ids == []
    assert history_manager.deleted_engram_ids == []

    engrams = {
        1: EngramInfo(id=1, type="WORKING", lifespan=4, age=0, fire_count=0, outgoings=[], incoming=[]),
        2: EngramInfo(id=2, type="SHORTTERM", lifespan=5, age=0, fire_count=0, outgoings=[], incoming=[]),
        3: EngramInfo(id=3, type="LONGTERM", lifespan=6, age=0, fire_count=0, outgoings=[], incoming=[]),
    }
    history_manager.add_summary(EngramsInfo(engrams=engrams, edges={}, working=[1], shortterm=[2], longterm=[3]))
    assert len(history_manager) == 1
    assert history_manager.timestep == 1
    assert history_manager.engram_ids == [1, 2, 3]
    assert history_manager.alive_engram_ids == [1, 2, 3]
    assert history_manager.deleted_engram_ids == []
    assert history_manager.engram_fire_counts == {}
    assert history_manager.engram_lastest_alive_timestep == {1: 0, 2: 0, 3: 0}
    assert history_manager.latest_engram_infos == {1: engrams[1], 2: engrams[2], 3: engrams[3]}

    engrams2 = {
        2: EngramInfo(id=2, type="SHORTTERM", lifespan=4, age=0, fire_count=0, outgoings=[], incoming=[]),
        3: EngramInfo(id=3, type="LONGTERM", lifespan=8, age=0, fire_count=1, outgoings=[], incoming=[]),
        4: EngramInfo(id=4, type="LONGTERM", lifespan=6, age=0, fire_count=0, outgoings=[], incoming=[]),
    }
    history_manager.add_summary(EngramsInfo(engrams=engrams2, edges={}, working=[], shortterm=[2], longterm=[3, 4]))
    assert len(history_manager) == 2
    assert history_manager.timestep == 2
    assert history_manager.engram_ids == [1, 2, 3, 4]
    assert history_manager.alive_engram_ids == [2, 3, 4]
    assert history_manager.deleted_engram_ids == [1]
    assert history_manager.engram_firing_times == {3: [1]}
    assert history_manager.engram_firings == {3: [Firing(timestep=1, engram_id=3, lifespan_gain=3.0)]}
    assert history_manager.engram_fire_counts == {3: 1}
    assert history_manager.engram_lastest_alive_timestep == {1: 0, 2: 1, 3: 1, 4: 1}
    assert history_manager.latest_engram_infos == {1: engrams[1], 2: engrams2[2], 3: engrams2[3], 4: engrams2[4]}

    engram_history = history_manager.inspect(1)
    assert engram_history.id == 1
    assert engram_history.creation_timestep == 0
    assert engram_history.deletion_timestep == 1
    assert engram_history.duration == 1
    assert engram_history.firing_times == []
    assert engram_history.firings == []



from setuptools import find_packages, setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="memoria-pytorch",
    version="1.1.0",
    description="Memoria is a human-inspired memory architecture for neural networks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.10",
    install_requires=["torch"],
    url="https://github.com/cosmoquester/memoria.git",
    author="Park Sangjun",
    keywords=["memoria", "hebbian", "memory", "transformer"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    packages=find_packages(exclude=["tests", "experiment"]),
)



from dataclasses import dataclass
from typing import Literal, Optional, Tuple


@dataclass(slots=True, frozen=True)
class EngramConnection:
    """Data structure for engram connections."""

    #: Source engram ID.
    source_id: int
    #: Target engram ID.
    target_id: int
    #: Connection weight (Probability).
    weight: float
    #: Cofire count.
    cofire_count: int


@dataclass(slots=True, frozen=True)
class EngramInfo:
    """Data structure for engram information."""

    #: Engram ID.
    id: int
    #: Engram type.
    type: Literal["WORKING", "SHORTTERM", "LONGTERM"]
    #: Lifetime of the engram.
    lifespan: int
    #: The age of the engram.
    age: Optional[int]
    #: Fire count of the engram.
    fire_count: int
    #: The outgoing edges of the engram.
    outgoings: Tuple[EngramConnection]
    #: The incoming edges of the engram.
    incoming: Tuple[EngramConnection]

    @property
    def cofire_counts(self) -> dict[int, int]:
        """Get the cofire counts of the engram."""
        return {edge.target_id: edge.cofire_count for edge in self.outgoings}


@dataclass(slots=True, frozen=True)
class EngramsInfo:
    """Data structure for engrams information."""

    #: Engram ID to EngramInfo mapping.
    engrams: dict[int, EngramInfo]
    #: All engram connections mapping from source and target engram IDs.
    edges: dict[Tuple[int, int], EngramConnection]
    #: Working memory engram IDs.
    working: Tuple[int]
    #: Short-term memory engram IDs.
    shortterm: Tuple[int]
    #: Long-term memory engram IDs.
    longterm: Tuple[int]


@dataclass(slots=True, frozen=True)
class Firing:
    """Data structure for firing information."""

    #: Firing timestep.
    timestep: int
    #: Engram ID.
    engram_id: int
    #: Lifespan Gain.
    lifespan_gain: float


@dataclass(slots=True, frozen=True)
class EngramHistory:
    """Historical information of an engram."""

    #: Engram ID.
    id: int
    #: Creation time of the engram.
    creation_timestep: int
    #: Deletion time of the engram.
    deletion_timestep: Optional[int]
    #: Duration of the engram.
    duration: Optional[int]
    #: Firing times of the engram.
    firing_times: list[int]
    #: Firing information of the engram.
    firings: list[Firing]
    #: Summaries of the engram.
    summaries: list[EngramsInfo]
